#!/usr/bin/perl
# Name: Web Quake 3 Arena Management CGI
# Description: Perl CGI interface to administer quake 3 servers and players on them
# Version: 2.0
# Authors: Jason Jorgensen <jasonj@innominatus.com>
# Copyright (C) 2003  Jason Jorgensen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################

use CGI qw/:standard :html3 :Carp/;
use strict;
use Socket;
use IO::Socket;
use IO::Select;
use Sys::Hostname;
use BSU::DynamicTemplate;
use CGI::Session::DB_File;
use DBI;
use DBIx::Broker;

my $programname = "Quake 3 Server Web Admin";

my $debug = 0;
my $showbanned = 1;
my $timeout = 1;
my $pausetime = 0;
my $script = $ENV{'SCRIPT_NAME'};
my $serverstemplate = 'templates/servers.html';
my $rconservertemplate = 'templates/rconserver.html';
my $servertemplate = 'templates/server.html';
my $aliastemplate = 'templates/aliases.html';
my $serversconfig = 'q3servers.cgi';
my $serversconfigdelimiter = '::';
my $navigation; # variable used to call different screens.
my $apacheuser = $ENV{'REMOTE_USER'}; # Apache authenticated user

my $dbuser = 'q3admin';
my $dbpassword = '7rackFR@GGERS';
my $dbdatabase = 'Q3Admin';
my $dbhost = 'localhost';
my $dbtype = 'mysql';
my $dbport = '3306';
my $dbtable = 'Aliases';
my $seentable = 'Seen';
my $freqcommandstable = 'FreqCommands';

my $socket;
my $s = IO::Select->new();
      

## If you want q3admin to record names and ip's set $q3record to 1
my $aliasdb = eval {DBIx::Broker->new($dbtype, $dbdatabase, $dbhost, $dbport, $dbuser, $dbpassword)};
if ($debug) { $aliasdb->debug_on; }

my $startform = start_form();
chomp $startform;

my @servers;


## Retrieve session id either from CGI variables, a cookie, or make it undefined
my $sessionid = param("webq3sid") || cookie('webq3sid') || undef;

## Regardless of the sessionid, start up a new session
my $session = new CGI::Session::DB_File($sessionid, {FileName=>'/tmp/webq3admin_sessions.db', LockDirectory=>'/tmp'});
if (! $session->param('navigation')) { $session = undef; $session = new CGI::Session::DB_File(undef, {FileName=>'/tmp/webq3admin_sessions.db', LockDirectory=>'/tmp'}); }

## Pass a cookie to the users browser with the current session id        
my $cookie = cookie(-name=>'webq3sid', -value=>$session->id, -expires=>"+3d");
                        
print header(-cookie=>$cookie);
&read_config(); ## Read configuration file that contains all the quake3 servers

if (! $aliasdb) { print "Can not connect with $programname database"; exit; }
print "DEBUG: sessionid: $sessionid<br>\n" if $debug;

my $counter = param("counter") || $session->param('counter');
if ($counter) {
  my $server = $servers[$counter-1];
  &q3open($server->{'host'}, $server->{'port'}, $timeout);
}

## Actions
my $action = param("action") || "";

if ($action eq 'ban') { 
  my $number = param("number");
  my $name = param("name");
  my $ip = param("ip");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "BAN/KICKING NUMBER $number $name $ip<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'wrconpassword'} g_filterBan 1");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} addip $ip");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} clientkick $number");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'kick') { 
  my $number = param("number");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "KICKING NUMBER $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} clientkick $number");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pbban') { 
  my $number = param("pbnumber");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "PB BANNING NUMBER $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_ban $number $name-PermBan $name-PermBan");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pbkick') { 
  my $number = param("pbnumber");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "KICKING NUMBER $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_kick $number 5 5MinuteKick 5MinuteKick");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pwr0') { 
  my $number = param("pbnumber");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "POWER LEVEL 0 (No Power) to pb player $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_power $number 0");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pwr1') { 
  my $number = param("pbnumber");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "POWER LEVEL 1 (Defualt) to pb player $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_power $number 1");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pwr25') { 
  my $number = param("pbnumber");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "POWER LEVEL 25 to pb player $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_power $number 25");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pwr50') { 
  my $number = param("pbnumber");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "POWER LEVEL 50 to pb player $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_power $number 50");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pwr75') { 
  my $number = param("pbnumber");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "POWER LEVEL 75 to pb player $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_power $number 75");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pwr100') { 
  my $number = param("pbnumber");
  my $name = param("name");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "POWER LEVEL 100 (Deputy) to pb player $number $name<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_power $number 100");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'unban') { 
  my $banip = param("banip");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "UNBANING IP $banip<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} removeip $banip");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'pbunban') { 
  my $guid = param("guid");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "UNBANING GUID $guid<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_unbanguid $guid");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
}

if ($action eq 'command') { 
  my $command = param("command");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "Sending Command: $command<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} $command");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
  #sleep 1;
}

if ($action eq 'maprestart') { 
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "RESTARTING SERVER $server->{'hostname'} <br><br><br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} map_restart\n");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
  #sleep 1;
}

if ($action eq 'rawcommand') { 
  my $command = param("rawcommand");
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "Sending Raw Command: $command<br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "$command");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
  #sleep 1;
}

if ($action eq 'maprestart') { 
  my $counter = param("counter") || $session->param('counter');
  
  my $server = $servers[$counter-1];
  print "RESTARTING SERVER $server->{'hostname'} <br><br><br>\n";
  print "Server Response:<br>";
  my @array = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} map_restart\n");
  foreach my $line (@{$array[0]}) { print $line."<br>\n"; }
  print "<hr><br>";
  #sleep 1;
}

print "DEBUG: apacheuser: $apacheuser<br>\n" if $debug;

## Navigation/Screens
print "DEBUG: param navigation: ".param("navigation")."<br>\n" if $debug;
$navigation = param("navigation") || $session->param('navigation') || 'servers';
print "DEBUG: session navigation: ".$session->param('navigation')."<br>\n" if $debug;
print "DEBUG: navigation: $navigation<br>\n" if $debug;

if ($navigation eq 'server') { $session->param(-name=>'navigation', -value=>$navigation); &display_server(); }
elsif ($navigation eq 'servers') { $session->param(-name=>'navigation', -value=>$navigation); &display_servers(); }
elsif ($navigation eq 'aliases') { &display_aliases(); }
else { &display_servers(); }

## Close out session so that DBFile is written
print "DEBUG: navigation: $navigation<br>\n" if $debug;
print "DEBUG: closing session<br>\n" if $debug;
$session->close;

$aliasdb->disconnect();

#&q3close();

exit;


## Sub Routines
sub display_aliases {
  my @aliases;
  my $ip = param("ip");
  my $name = param("name");
  #$name =~ s/\'/\\'/g;
  my $guid = param("guid");
  my $playerfont = param("playerfont") || $session->param('playerfont') || 'Text'; # user selection to calculate $refresh
  my $search = param("aliassearch");

  if ($search) {
    my @search = split //, $search;
    foreach my $char (@search) {
      $char =~ s/\./\\./g;
      $char =~ s/\(/\\(/g;
      $char =~ s/\)/\\)/g;
      $char =~ s/\*/\\*/g;
      $char =~ s/\+/\\+/g;
      $char =~ s/\[/\\[/g;
      $char =~ s/\]/\\]/g;
      $char =~ s/\{/\\{/g;
      $char =~ s/\}/\\}/g;
    }
    $guid = $search;
    $search = '.*'.(join '.*', @search).'.*';
    print "DEBUG: search string: $search<br>\n" if $debug;
    $ip = $name = $search;
  }

  print "DEBUG: ip: $ip<BR>\n" if $debug;
  print "DEBUG: name: $name<BR>\n" if $debug;
  print "DEBUG: guid: $guid<BR>\n" if $debug;

  my $aliasentry;
  my @aliasdata;
  my $guidentry;
  my @guiddata;
  my @fields  =  ( "Name", "IP", "GUID", "LastSeen", "FirstSeen", "TimesSeen", "EstTimeInGame" );

  if (($ip !~ /\s+/) || ($name !~ /\s+/) || ($guid !~ /\s+/)) {
    if (! $guid) { $guid = "NO MATCH"; }
    if (! $name) { $name = "NO MATCH"; }
    if (! $ip) { $ip = "NO MATCH"; }
    if ($search) {
      @aliases = $aliasdb->select(\@fields, $dbtable, "WHERE GUID REGEXP '$guid' OR IP REGEXP '$ip' OR Name REGEXP '$name'", 1);    
    }
    else {
      @aliases = $aliasdb->select(\@fields, $dbtable, "WHERE GUID='$guid' OR IP='$ip' OR Name='$name'", 1);    
    }
  }

  foreach my $alias (@aliases) {
    my $lastseen = localtime($alias->{'LastSeen'});
    my $firstseen = localtime($alias->{'FirstSeen'});
    if ($firstseen eq 'Wed Dec 31 18:00:00 1969') { $firstseen = "N/A"; }
    if ($lastseen eq 'Wed Dec 31 18:00:00 1969') { $lastseen = "N/A"; }
    $aliasentry = {
                    name => &prettyname($alias->{'Name'}, $playerfont),
                    namelink => "$script?navigation=aliases&name=$alias->{'Name'}",
                    ip   => $alias->{'IP'},
                    iplink => "$script?navigation=aliases&ip=$alias->{'IP'}",
                    guid   => $alias->{'GUID'},
                    guidlink => "$script?navigation=aliases&guid=$alias->{'GUID'}",
                    lastseen => "$lastseen",
                    firstseen => "$firstseen",
                    timesseen => "$alias->{'TimesSeen'}",
                  };    
    push @aliasdata, $aliasentry;
  }
  my $aliastemplate = new BSU::DynamicTemplate "$aliastemplate";
  $aliastemplate->set_loop_variables(aliastable => \@aliasdata);
  if ($guid ne "NO MATCH") {
    my @fields  =  ( "GUID", "LastSeen", "FirstSeen", "TimesSeen");
    my $seeninfo = $aliasdb->select_one_row(\@fields, $seentable, 'WHERE GUID=\''.$guid.'\'',1);
    my $slastseen = localtime($seeninfo->{'LastSeen'});
    my $sfirstseen = localtime($seeninfo->{'FirstSeen'});
    $guidentry = {
                  guid => $seeninfo->{'GUID'},
                  firstseen => $sfirstseen,
                  lastseen => $slastseen,
                  timesseen => $seeninfo->{'TimesSeen'},
                 };
    push @guiddata, $guidentry;
  $aliastemplate->set_loop_variables(guidtable => \@guiddata);
  }
  $aliastemplate->set(
                        searchfield => $startform.textfield(-name=>'aliassearch',
                                          -default=>"",
                                          -override=>1,
                                          -size=>'15',
                                          -maxlength=>'256').submit(-name=>'Search').hidden(-name=>'navigation',-override=>1,-value=>'aliases').hidden(-name=>'action',-override=>1, -value=>'aliassearch').end_form,
                      );
  print $aliastemplate->render();  
}

sub display_server {
  my @playersdata;
  my $counter = param("counter") || $session->param('counter');
  my $autorefresh = param("autorefresh") || $session->param('autorefresh') || 'Never'; # user selection to calculate $refresh
  my $refresh;  # used to set the auto refresh meta tag
  my $playerfont = param("playerfont") || $session->param('playerfont') || 'Text'; # user selection to calculate $refresh
  my $fontsize;
  
  my $server = $servers[$counter-1];
  if ($server->{'access'}{$apacheuser}) {
  #my $socket = &q3open($server->{'host'}, $server->{'port'}, $timeout);
  my @temparray;
  my $response;
  my $svplist; # punk buster server style playerlist
  my $game;
  my $playerdata;
  my %pbdata; #decoded playerlist

  ## Get the current info for just this one selected server
  my $serverdata;
  $response = &q3command($server->{'host'}, $server->{'port'}, $timeout, "getinfo");
  my @serverarray = split /\\/, @$response[0];
  splice (@serverarray,0,1);

  ## Get the current list of players for just this one selected server
  if ($server->{'rconpassword'}) {
    $response = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} status");
    #sleep 2;
    $svplist = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_plist");
  }
  else {
    $response = &q3command($server->{'host'}, $server->{'port'}, $timeout, "getstatus");
  }
  if (! $response) { &problemdown(); }
  my @players = @$response;
  if ($svplist) { %pbdata = &svplistdecode($svplist); }
  if ($server->{'rconpassword'}) {
    splice (@players,0,2);
  }
  else {
    splice (@players,0,1);
  }
  foreach my $player (@players) {
    if (($player =~ /\S+/) && ($player !~ /-----------/) && ($player !~ /num\s+score\s+ping\s+name\s+lastmsg\s+address\s+qport\s+rate/)){
      my $playerdata;
      my $ip;
      my $port;

      my @splitplayer = split /\s+/, $player;
      my @playername;
      my $playername;
      if ($server->{'rconpassword'}) {
        splice (@splitplayer,0,1);
        @playername = splice (@splitplayer,3,$#splitplayer-6,'UNUSED');
        $playername = join ' ', @playername;
#        my $pbplayername = $playername;
#        $playername =~ s/\s+/\ /g;
#         print "DEBUG: GUID for $playername: $pbdata{$pbplayername}[1]<br>\n" if $debug;
#        $playername =~ s/\s+\^7$//g;
        ($ip, $port) = split /:/, $splitplayer[5];
        if ($splitplayer[2] eq '999') {$splitplayer[2] = "<font color=red>999</font>"; }
        my $correctedname = $playername;
        print "DEBUG: name: $correctedname<br>\n" if $debug;
        $correctedname =~ s/\"//g;
        $correctedname =~ s/\'//g;
#        $correctedname =~ s/\^/\^/g;
#        $correctedname =~ s/\$/\$/g;
#        $correctedname =~ s/\&/\&/g;
        print "DEBUG: stripped name: $correctedname<br>\n" if $debug;
        if (($correctedname) && ($ip) && ($pbdata{$playername}[1])) {
          &record($correctedname,$ip,$pbdata{$playername}[1]);
        }
        my @fields  =  ( "LastSeen", "FirstSeen", "TimesSeen", "EstTimeInGame" );
        my $searchname = $playername;
        $searchname =~ s/\'/\\\'/g;
        my $userinfo = $aliasdb->select_one_row(\@fields, $dbtable, 'WHERE Name=\''.$searchname.'\' AND IP=\''.$ip.'\' AND GUID=\''.$pbdata{$playername}[1].'\'',1);
        my @fields  =  ( "LastSeen", "FirstSeen", "TimesSeen");
        my $seeninfo = $aliasdb->select_one_row(\@fields, $seentable, 'WHERE GUID=\''.$pbdata{$playername}[1].'\'',1);
        my $lastseen = localtime($userinfo->{'LastSeen'});
        my $firstseen = localtime($userinfo->{'FirstSeen'});
        my $slastseen = localtime($seeninfo->{'LastSeen'});
        my $sfirstseen = localtime($seeninfo->{'FirstSeen'});

        $firstseen =~ s/Wed Dec 31 18\:00\:00 1969/N\/A/g;
        $sfirstseen =~ s/Wed Dec 31 18\:00\:00 1969/N\/A/g;
        $lastseen =~ s/Wed Dec 31 18\:00\:00 1969/N\/A/g;
        $slastseen =~ s/Wed Dec 31 18\:00\:00 1969/N\/A/g;
        $playerdata = {
                     number  => $splitplayer[0],
                     score   => $splitplayer[1],
                     ping    => $splitplayer[2],
                     name    => $playername,
                     lastmsg => $splitplayer[4],
                     ip      => $ip,
                     port    => $port,
                     qport   => $splitplayer[6],
                     rate    => $splitplayer[7],
                     lastseen => "$lastseen",
                     firstseen => "$firstseen",
                     userlastseen => "$slastseen",
                     userfirstseen => "$sfirstseen",
                     timesseen => "$userinfo->{'TimesSeen'}",
                     usertimesseen => "$seeninfo->{'TimesSeen'}",
                     pbnumber=> $pbdata{$playername}[0],
                     guid    => $pbdata{$playername}[1],
                     pbip    => $pbdata{$playername}[2],
                     status  => $pbdata{$playername}[3],
                     power   => $pbdata{$playername}[4],
                     authnum => $pbdata{$playername}[5],
                     ssnum   => $pbdata{$playername}[6],
                     os      => $pbdata{$playername}[7],
                     validity=> $pbdata{$playername}[9],
                    };    
       }
       else {
         @playername = splice (@splitplayer,2,$#splitplayer,'UNUSED');
         $playername = join ' ', @playername;
         $playerdata = {
                     score   => $splitplayer[0],
                     ping    => $splitplayer[1],
                     name    => $playername,
                    };    
      }
#print "DEBUG: 7&7: 7&7 ^7: $pbdata{'7&7 ^7'}[1]<br>\n";
#print "DEBUG: 7&7: 7&7^7: $pbdata{'7&7^7'}[1]<br>\n";
#print "DEBUG: 7&7: 7&7: $pbdata{'7&7'}[1]<br>\n";

      print "DEBUG: player line: $player<br>\n" if $debug;
      print "DEBUG: Playername: @playername<br>\n" if $debug;
      print "DEBUG: splitplayer[0]: $splitplayer[0]<br>\n" if $debug;
      print "DEBUG: splitplayer[1]: $splitplayer[1]<br>\n" if $debug;
      print "DEBUG: splitplayer[2]: $splitplayer[2]<br>\n" if $debug;
      print "DEBUG: splitplayer[3]: $splitplayer[3]<br>\n" if $debug;
      print "DEBUG: splitplayer[4]: $splitplayer[4]<br>\n" if $debug;
      print "DEBUG: splitplayer[5]: $splitplayer[5]<br>\n" if $debug;
      print "DEBUG: splitplayer[6]: $splitplayer[6]<br>\n" if $debug;
      print "DEBUG: splitplayer[7]: $splitplayer[7]<br>\n" if $debug;
      print "DEBUG: guid: $playerdata->{'guid'}<br>\n" if $debug;
      print "DEBUG: os: $playerdata->{'os'}<br>\n" if $debug;
      print "DEBUG: pbnumber: $playerdata->{'pbnumber'}<br>\n" if $debug;

      my $correctedname = $playerdata->{'name'};
      print "DEBUG: name: $correctedname<br>\n" if $debug;
      $correctedname =~ s/\"//g;
      $correctedname =~ s/\'//g;

      ## convert player name into images
      $playerdata->{'name'} = &prettyname($playerdata->{'name'}, $playerfont);


      if ($server->{'rconpassword'}) {
        $playerdata->{'ipaliaslink'} = "$script?navigation=aliases&ip=$ip";
        $playerdata->{'actionmenu'} = popup_menu('action',['kick','ban','pbkick','pbban','pwr0','pwr1','pwr25','pwr50','pwr75','pwr100'],'kick').submit(-name=>'Go');
        $playerdata->{'formstart'} = $startform.hidden(-name=>'number',-override=>1,-value=>$playerdata->{'number'}).hidden(-name=>'pbnumber',-override=>1,-value=>$playerdata->{'pbnumber'}).hidden(-name=>'ip',-override=>1,-value=>$playerdata->{'guid'});
        $playerdata->{'formend'} = end_form;
      }
      $playerdata->{'namealiaslink'} = "$script?navigation=aliases&name=$correctedname";
      $playerdata->{'aliaslink'} = "$script?navigation=aliases&name=$correctedname&ip=$ip&guid=$playerdata->{'guid'}";

      print "$player<br>\n" if $debug;
      if (($playerdata-{'ping'}) && ($playerdata->{'name'})) {
        push @playersdata, $playerdata;
      }
    }
  }
  

  ## Display list of frequently used commands
  my @freqcommandsdata;
  my @fields  =  ( "CommandID", "Command", "Description");
  my @freqcommands = $aliasdb->select(\@fields, $freqcommandstable, "", 1);    

  for(my $c=0;$c<=$#freqcommands;$c++) {

    my $commandentry = {
                    num => $c,
                    command => $freqcommands[$c]->{'Command'},
                    description => "$freqcommands[$c]->{'Command'} - $freqcommands[$c]->{'Description'}",
                    };
    push @freqcommandsdata, $commandentry;
  }


  ## Get the current variables for this one selected server
  my @variablesdata;
  $response = &q3command($server->{'host'}, $server->{'port'}, $timeout, "getstatus");
  if (! $response) { &problemdown(); }
  my @variables = split /\\/, @$response[0];
  splice (@variables,0,1);
  for (my $c = 0; $c < $#variables; $c = $c+2) {
    my $variabledata;
    $variabledata->{'name'} = $variables[$c];
    $variabledata->{'value'} = $variables[($c+1)];
    $variabledata->{'field'} = textfield(-name=>var-$variables[$c],
                                          -default=>"",
                                          -override=>1,
                                          -size=>(length($variables[($c+1)]) + 1),
                                          -maxlength=>(length($variables[($c+1)]) * 2));
    
    print "$variables[$c] => $variables[($c+1)]<br>\n" if $debug;
    push @variablesdata, $variabledata;
  }

  
  ## Get currently banned IP's and players
  my @bannedipsdata;
  if (($showbanned) && ($server->{'rconpassword'})){
    #sleep $pausetime;
    my @bannediparray = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} g_banIps");
    my ($bannedips) = (${$bannediparray[0]}[0] =~ /\:\"(.*)\^7\"\ default/);
    my @bannedips = split / /, $bannedips || undef;
    foreach my $banip (@bannedips) { 
      my $bannedipdata;
      $bannedipdata->{'banned'} = $banip;
      $bannedipdata->{'unbanbutton'} = $startform.submit(-name=>'Unban').hidden(-name=>'action',-override=>1,-value=>'unban').hidden(-name=>'banip',-override=>1,-value=>$banip).end_form; 
      $bannedipdata->{'info'} = "Quake3 IP BAN";
      $bannedipdata->{'aliaslink'} = "$script?navigation=aliases&ip=$banip";
      push @bannedipsdata, $bannedipdata;
    }
    my $bannedplayerarray = &q3command($server->{'host'}, $server->{'port'}, $timeout, "rcon $server->{'rconpassword'} pb_sv_banlist");
    my @bannedplayerarray = @{$bannedplayerarray};
    my @bannedplayers;
    my @bannedplayersinfo;
    foreach my $line (@bannedplayerarray) {
      if ($line !~ /End of Ban List/) {
        print "DEBUG: bannedplayerarray: $line<br>\n" if $debug;
        my ($bannedplayer) = ($line =~ /\b(.{32,32}\b)/);
        my ($bannedplayerinfo) = ($line =~ /\}\ (.*)\n/);
        if ($bannedplayer !~ /\s/) {
          push @bannedplayers, $bannedplayer;
          push @bannedplayersinfo, $bannedplayerinfo;
          print "DEBUG: bannedplayer: $bannedplayer<br>\n" if $debug;
          print "DEBUG: bannedplayerinfo: $bannedplayerinfo<br>\n" if $debug;
        } 
      }
    }
    for(my $c=0;$c<=$#bannedplayers;$c++) {
      print "DEBUG: bannedplayer: $bannedplayers[$c]<br>\n" if $debug;
      my $bannedipdata;
      $bannedipdata->{'banned'} = $bannedplayers[$c];
      $bannedipdata->{'unbanbutton'} = $startform.submit(-name=>'Unban').hidden(-name=>'action',-override=>1,-value=>'pbunban').hidden(-name=>'guid',-override=>1,-value=>$bannedplayers[$c]).end_form; 
      $bannedipdata->{'info'} = $bannedplayersinfo[$c];
      $bannedipdata->{'aliaslink'} = "$script?navigation=aliases&guid=$bannedplayers[$c]";
      push @bannedipsdata, $bannedipdata;
    }
  }

  #&q3close($socket);
  #if (! $serverdata->{'game'}) { $serverdata->{'game'} = &gametype($serverdata->{'gametype'}); }
  
  ## Calculate the autorefresh time
  if ($autorefresh eq "Never") { $refresh = ""; }
  elsif ($autorefresh eq "15 Seconds") { $refresh = "<META HTTP-EQUIV=\"REFRESH\" content=\"15\">"; }
  elsif ($autorefresh eq "30 Seconds") { $refresh = "<META HTTP-EQUIV=\"REFRESH\" content=\"30\">"; }
  elsif ($autorefresh eq "1 Minute") { $refresh = "<META HTTP-EQUIV=\"REFRESH\" content=\"60\">"; }
  elsif ($autorefresh eq "5 Minutes") { $refresh = "<META HTTP-EQUIV=\"REFRESH\" content=\"300\">"; }
  elsif ($autorefresh eq "10 Minutes") { $refresh = "<META HTTP-EQUIV=\"REFRESH\" content=\"600\">"; }
  else { $refresh = ""; }

  
 my $memememe;
  if ($server->{'rconpassword'}) {
    $memememe = $rconservertemplate;
  }
  else {
    $memememe = $servertemplate;
  }
  my $servertemplate = new BSU::DynamicTemplate "$memememe";
  for (my $c = 0; $c < $#serverarray; $c = $c+2) {
    $servertemplate->set($serverarray[$c] => $serverarray[($c+1)]);
    if ($serverarray[$c] eq 'gametype' && ! grep (/^game$/, @serverarray)) { 
      $servertemplate->set(game =>  &gametype($serverarray[($c+1)])); 
      print "DEBUG: game type: $serverarray[($c+1)]<br>\n" if $debug;
      print "DEBUG: game type: $game<br>\n" if $debug;
    }
  }
  $servertemplate->set(
                        startform => $startform,
                        endform => end_form(),
                        serversbutton => $startform.submit(-name=>'All Servers Page').hidden(-name=>'navigation',-override=>1, -value=>'servers').end_form,
                        refreshbutton => $startform.submit(-name=>'Refresh Page').end_form,
                        autorefreshpopup => $startform.popup_menu(-name=> 'autorefresh', -values=>['Never','15 Seconds','30 Seconds','1 Minute','5 Minutes','10 Minutes'], -default=>"$autorefresh").submit(-name=>'Set Auto Refresh').end_form,
                        playerfont => $startform.popup_menu(-name=> 'playerfont', -values=>['Text','Normal','Big','Huge'], -default=>"$playerfont").submit(-name=>'Set Font Type').end_form,
                        serverhidden => hidden(-name=>'counter',-override=>1, -value=>$counter),
                        refresh => $refresh,
                        host => $server->{'host'},
                        port => $server->{'port'},
                       );
  if ($server->{'rconpassword'}) {
    $servertemplate->set(
                        restartbutton => $startform.submit(-name=>'Map Restart').hidden(-name=>'action',-override=>1, -value=>'maprestart').end_form,
                        commandfield => start_form(-name=>'rconcommand').textfield(-name=>'command',
                                          -default=>"",
                                          -override=>1,
                                          -size=>'50',
                                          -maxlength=>'256').submit(-name=>'Send Command').hidden(-name=>'action',-override=>1, -value=>'command').end_form,
                        rawcommandfield => $startform.textfield(-name=>'rawcommand',
                                          -default=>"",
                                          -override=>1,
                                          -size=>'50',
                                          -maxlength=>'256').submit(-name=>'Send Raw Command').hidden(-name=>'action',-override=>1, -value=>'rawcommand').end_form,
                       aliaspage => "$script?navigation=aliases",
                      );
  }  
  $servertemplate->set_loop_variables(bannediptable => \@bannedipsdata);
  $servertemplate->set_loop_variables(playerstable => \@playersdata);
  $servertemplate->set_loop_variables(variablestable => \@variablesdata);
  $servertemplate->set_loop_variables(freqcommandstable => \@freqcommandsdata);

  ## Before we render and print the screen, store any useful variables in the session
  $session->param(-name=>'counter', -value=>$counter);
  $session->param(-name=>'autorefresh', -value=>$autorefresh);
  $session->param(-name=>'playerfont', -value=>$playerfont);

  ## Render and print the screen
  print $servertemplate->render();  
  }
  else {
    print "Access Denied<BR>\n";
    print $startform.submit(-name=>'All Servers Page').hidden(-name=>'navigation',-override=>1, -value=>'servers').end_form;
  }
}

sub display_servers {
  ## if no navigation variables are populated display list of servers to be managed
  my @serversdata;

  foreach my $server (@servers) {
    if ($server->{'access'}{$apacheuser}) {    
      my $serverdata;
      my @temparray;
      my $response = &q3commandsingle($server->{'host'}, $server->{'port'}, $timeout, "getinfo");
      @temparray = split /\\/, @$response[0];
      splice (@temparray,0,1);
      for (my $c = 0; $c < $#temparray; $c = $c+2) {
        print "$temparray[$c] => $temparray[($c+1)]<br>\n" if $debug;
        $serverdata->{$temparray[$c]} = $temparray[($c+1)];
      }
      if ($server->{'rconpassword'}) {
        $serverdata->{'button'} = $startform.submit(-name=>'Manage').hidden(-name=>'counter', -override=>1,-value=>$server->{'counter'}).hidden(-name=>'navigation',-override=>1,-value=>'server').end_form;
      }
      else {
        $serverdata->{'button'} = $startform.submit(-name=>'View').hidden(-name=>'counter', -override=>1,-value=>$server->{'counter'}).hidden(-name=>'navigation',-override=>1,-value=>'server').end_form;
      }
      $serverdata->{'counter'} = $server->{'counter'};
      $serverdata->{'host'} = $server->{'host'};
      $serverdata->{'port'} = $server->{'port'};
      if (! $serverdata->{'game'}) { $serverdata->{'game'} = &gametype($serverdata->{'gametype'}); }
      push @serversdata, $serverdata;
      #&q3close($socket);
    }
#    else { print "Access Denied<br>\n"; }
  }

  my $serverstemplate = new BSU::DynamicTemplate "$serverstemplate";
  $serverstemplate->set(
                        refreshbutton => $startform.submit(-name=>'Refresh Page').end_form,
                        startform => $startform,
                        user => $apacheuser,
                        endform => end_form(),
                       );
  $serverstemplate->set_loop_variables(serverstable => \@serversdata);
  print $serverstemplate->render();  
  
}

sub gametype {
  my $type = shift @_;
  
  if (! $type) { return "Unknown, Down, or Between Levels"; }

  if ($type == 0) { return "Free for All";}     # 0 - Free For All DM (FFA)  [default]
  if ($type == 1) { return "1v1 DeathMatch";}   # 1 - 1v1 DM
  if ($type == 2) { return "Single Player";}    # 2 - Single-Player
  if ($type == 3) { return "Team Deathmatch";}  # 3 - Team Deathmatch
  if ($type == 4) { return "Capture The Flag";} # 4 - Capture The Flag (CTF) Teamplay
  if ($type == 5) { return "Clan Arena";}       # 5 - Clan Arena (CA)

  return "Unknown";

}

sub htmlcolor {
  my $color = shift @_;

  $color = orc($color) % 8; # take the modulus of the ascii value of the character (thanks broo)

  if ($color == 0) { return "\#2F4F4F"; }
  elsif ($color == 1) { return "red"; }
  elsif ($color == 2) { return "green"; }
  elsif ($color == 3) { return "yellow"; }
  elsif ($color == 4) { return "blue"; }
  elsif ($color == 5) { return "lightblue"; }
  elsif ($color == 6) { return "violet"; }
  elsif ($color == 7) { return "white"; }
  elsif ($color == 8) { return "\#2F4F4F"; }
  elsif ($color == 9) { return "red"; }
}

sub prettyname {
  my $name = shift @_;
  my $font = shift @_;
  my $fontsize;

  $name =~ s/^\"//g;
  $name =~ s/\"$//g;
  my $colorless = $name;
  $colorless =~ s/\^\^/-DOUBLECARET-\^/g;
  $colorless =~ s/\^.//g;
  $colorless =~ s/-DOUBLECARET-/\^/g;
  if ($colorless eq "") { $name .= "<BLINK>!COLORNAME ABUSER!</BLINK>"; }
  print "DEBUG: colorless: $colorless<br>\n" if $debug;
  if ($font eq "Text") {
    $name = "<font color=white>$name";
    $name =~ s/\^\^/-DOUBLECARET-\^/g;
    $name =~ s/\^$/-ENDCARET-/g;
    $name =~ s/\^(.)/'^'.(ord($1) % 8)/eg;
    $name =~ s/-DOUBLECARET-/\^/g;
    $name =~ s/-ENDCARET-/\^/g;
    $name =~ s/\^0/\<\/font\>\<font color\=\#2F4F4F\>/g;
    $name =~ s/\^1/\<\/font\>\<font color\=red\>/g;
    $name =~ s/\^2/\<\/font\>\<font color\=green\>/g;
    $name =~ s/\^3/\<\/font\>\<font color\=yellow\>/g;
    $name =~ s/\^4/\<\/font\>\<font color\=blue\>/g;
    $name =~ s/\^5/\<\/font\>\<font color\=lightblue\>/g;
    $name =~ s/\^6/\<\/font\>\<font color\=violet\>/g;
    $name =~ s/\^7/\<\/font\>\<font color\=white\>/g;
    $name =~ s/\^8/\<\/font\>\<font color\=\#2F4F4F\>/g;
    $name =~ s/\^9/\<\/font\>\<font color\=red\>/g;
    $name .= "</font>";
  }
  else {
    if ($font eq "Normal") { $fontsize = ""; }
    elsif ($font eq "Big") { $fontsize = "-BIG"; }
    elsif ($font eq "Huge") { $fontsize = "-HUGE"; }
    else {$fontsize = ""; }
    my @name = split //, $name;
    $name = "";
    my $color = "white";
    for (my $c=0; $c <= $#name; $c++) {
      if ($name[$c] eq '^') {
        $name[$c+1] = ord($name[$c+1]) % 8;
        if ($name[$c+1] eq '0') { $color = "black"; $c++;}
        elsif ($name[$c+1] eq '1') { $color = "red"; $c++;}
        elsif ($name[$c+1] eq '2') { $color = "green"; $c++;}
        elsif ($name[$c+1] eq '3') { $color = "yellow"; $c++;}
        elsif ($name[$c+1] eq '4') { $color = "dblue"; $c++;}
        elsif ($name[$c+1] eq '5') { $color = "lblue"; $c++;}
        elsif ($name[$c+1] eq '6') { $color = "violet"; $c++;}
        elsif ($name[$c+1] eq '7') { $color = "white"; $c++;}
        elsif ($name[$c+1] eq '8') { $color = "black"; $c++;}
        elsif ($name[$c+1] eq '9') { $color = "red"; $c++;}
        else { $name .= "<img src=\"dec-images/$color-94$fontsize.png\">"; }
      }
      else { $name .= "<img src=\"dec-images/$color-".ord($name[$c])."$fontsize.png\">"; }
    }
  }
  return $name;
}


sub read_config {
  my $counter = 1;
  open CONFIG, "<$serversconfig";
  foreach my $line (<CONFIG>) {
    ## If the line in the config file matches our scheme then add it to the server hash
    if (($line =~ /^.*$serversconfigdelimiter\d*/) && ($line !~ /^#/)){
      my %access;
      (my $host, my $port, my $rconpassword, my $accessstring) = split $serversconfigdelimiter, $line;
      chomp ($host, $port, $rconpassword, $accessstring);
      my @accessarray = split /,/, $accessstring;
      foreach my $user (@accessarray) { $access{$user} = 1;}
      if (($host =~ /.+\.\w+/ || $host =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/) && $port =~ /\d+/) {
      my $temphash = { counter      => $counter,
                       host         => $host,
                       port         => $port,
		       rconpassword => $rconpassword,
                       access       => \%access,
		     };
      push @servers, $temphash;
      $counter++;
      }
    }
  }
  close CONFIG;
  if ($debug) {
    foreach my $server (@servers) {
      print "counter: $server->{'counter'}<br>\n";
      print "host: $server->{'host'}<BR>\n";
      print "port: $server->{'port'}<BR>\n";
      print "rconpassword: $server->{rconpassword}<BR>\n";
      my $user;
      foreach $user (keys %{$server->{'access'}}) {
        print "access user: $user<br>\n";
      }
    }
  }
}


sub problemdown {
  print <<END;
The server you are querying is either:<br>\n
- Down<br>\n
- Changing Maps<br>\n
- Unreachable<br>\n
<br>\n
This is a potential error received from the server:<br>
<PRE>
@_
</PRE>
<br>\n
<CENTER>Refresh this page in a few seconds to see if the server is responding yet</CENTER><br>\n
</BODY>
</HTML>
END

exit 0;
}

sub record {
  my $name = shift @_;
  my $searchname = $name;
  #$searchname =~ s/\'/\\\'/g;
  my $ip = shift @_;
  my $guid = shift @_;
  my $time = time;

  if (($ip) && ($name)) {
    print "DEBUG: $name count: ".$aliasdb->count($dbtable, 'WHERE Name=\''.$searchname.'\' AND IP=\''.$ip.'\' AND GUID=\''.$guid.'\'')."<br>\n" if $debug;
    print "DEBUG: \$aliasdb->select_one_row($dbtable, 'WHERE Name=\'$searchname\' AND IP=\'$ip\' AND GUID=\'$guid\'')<br>\n" if $debug;
    my @fields = ("Name", "IP", "GUID", "FirstSeen", "LastSeen", "EstTimeInGame", "TimesSeen");
    my $user = $aliasdb->select_one_row(\@fields, $dbtable, 'WHERE Name=\''.$searchname.'\' AND IP=\''.$ip.'\' AND GUID=\''.$guid.'\'',1);
    my $firstseen = $user->{'FirstSeen'} || $time;
    my $lastseen = $user->{'LastSeen'};
    my $esttime = $user->{'EstTimeInGame'};
     my $newesttime = ($esttime + ($time - $lastseen));
    my $newesttime = 0;
    my $seen = $user->{'TimesSeen'} || 0;
    my $newseen = $seen + 1;
    print "DEBUG: SEEN: $seen<br>\n" if $debug;
    print "DEBUG: SEEN: $newseen<br>\n" if $debug;
    my %alias = (
                 "Name",          $name,
                 "IP",            $ip,
                 "GUID",          $guid,
                 "FirstSeen",     $firstseen,
                 "LastSeen",      $time,
                 "EstTimeInGame", $newesttime,
                 "TimesSeen",     $newseen,
                );
    if (! $user->{'Name'}) {
      print "DEBUG: INSERTING user record for: $searchname<br>\n" if $debug;
      $aliasdb->insert($dbtable, \%alias);
    }
    else {
      print "DEBUG: UPDATING user record for: $searchname<br>\n" if $debug;
      $aliasdb->update($dbtable, \%alias, 'WHERE Name=\''.$searchname.'\' AND IP=\''.$ip.'\' AND GUID=\''.$guid.'\'');
    }

    my @fields = ("GUID", "FirstSeen", "LastSeen", "TimesSeen");
    my $userseen = $aliasdb->select_one_row(\@fields, $seentable, 'WHERE GUID=\''.$guid.'\'',1);
    foreach my $key (keys(%{$userseen})) {
      print "DEBUG: seenkey: $key: $userseen->{$key}<br>\n" if $debug;
    }
    my $sfirstseen = $userseen->{'FirstSeen'} || $time;
    my $slastseen = $userseen->{'LastSeen'};
    my $sesttime = $userseen->{'EstTimeInGame'};
     my $snewesttime = ($esttime + ($time - $lastseen));
    my $snewesttime = 0;
    my $sseen = $userseen->{'TimesSeen'} || 0;
    my $snewseen = $sseen + 1;
    print "DEBUG: SEENSEEN: $sseen<br>\n" if $debug;
    print "DEBUG: SEENSEEN: $snewseen<br>\n" if $debug;
    my %seen = (
                 "GUID",          $guid,
                 "FirstSeen",     $sfirstseen,
                 "LastSeen",      $time,
                 "TimesSeen",     $snewseen,
                );
    if (! $userseen->{'GUID'}) {
      print "DEBUG: INSERTING seen record for: $searchname<br>\n" if $debug;
      $aliasdb->insert($seentable, \%seen);
    }
    else {
      print "DEBUG: UPDATING seen record for: $searchname<br>\n" if $debug;
      $aliasdb->update($seentable, \%seen, 'WHERE GUID=\''.$guid.'\'');
    }
  }
}

sub svplistdecode {
  my $list = shift @_;

  my @rawlist;
  my %cleanlist;

  if ($list) {
  print "DEBUG: server playerlist decode<br>\n" if $debug;
  @rawlist = @{$list};
  #splice (@rawlist,0,1);
  if ($#rawlist > 0) { splice (@rawlist,$#rawlist,1); }
  print "<table width=\"100%\">\n" if $debug;
  print "<tr><td>[Slot #]</td><td>[GUID]</td><td>[Address]</td><td>[Status]</td><td>[Power]</td><td>[Auth Rate]</td><td>[Recent SS]</td><td>[O/S]</td><td>[Name]</td></tr>\n" if $debug;
  foreach my $entry (@rawlist) {
    print "DEBUG: rawentry: $entry<br>\n" if $debug;
    my @rawentry;

    ($rawentry[9]) = ($entry =~ /\b[\w|\d]{32,32}\b\((\S*)\)/); # Validity
    ($rawentry[8]) = ($entry =~ /\(\w\)\ (.*)\n/); # name
    ($rawentry[7]) = ($entry =~ /\((\w)\)/); # OS
    ($rawentry[6]) = ($entry =~ /\ (\d+)\ \(.\)/); # Recent SS Number
    ($rawentry[5]) = ($entry =~ /\ (\d+\.\d+)\ \d+\ \(.\)/); # Auth Number
    ($rawentry[4]) = ($entry =~ /\ (\d+)\ \d+\.\d+\ \d+\ \(.\)/); # Power Level
    ($rawentry[3]) = ($entry =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\:\d+\ (\S*)\ /); # Status
    ($rawentry[2]) = ($entry =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\:\d+)/); # IP:PORT
    ($rawentry[1]) = ($entry =~ /\b([\w|\d]{32,32})\b/); # GUID
    ($rawentry[0]) = ($entry =~ /\ (\d+)\ /); # Slot
    $rawentry[8] .= "^7";
    print "<tr>\n" if $debug;
    $cleanlist{"$rawentry[8]"} = [ @rawentry ];
    foreach my $element (@rawentry) { print "<td>$element</td>" if $debug; }
    print "</tr>\n" if $debug;
  }
  print "</table>\n" if $debug;
  }

  return %cleanlist;
}

sub q3open {
  my $host = shift @_;
  my $port = shift @_;
  my $timeout = shift @_;

  $socket = IO::Socket::INET->new(PeerAddr => $host,
                                     PeerPort => $port,
                                     Proto    => 'udp',
                                    );
#                                     Timeout  => $timeout,
#                                     Reuse    => 1,
#                                     Listen   => 10,
  if (! $socket) {
   print "Cant talk to server at '$host:$port'\n";
   die "Cant talk to server at '$host:$port'\n";
   exit;
  }
  #$socket->autoflush(1);
  $s->add($socket);
	      
  return $socket;
}

sub q3close {
  my $socket = shift @_;

  close $socket;
}


sub q3command {
  my $host = shift @_ || "";
  my $port = shift @_ || "";
  my $timeout = shift @_ || "";
  my $msg = shift @_ || "";
  my $buffer;
  my $code = "\377\377\377\377";
  $msg = $code.$msg;
 
  my $socket;
  my $accumbuffer = "";


  my @wready = $s->can_write($timeout);
  print "DEBUG: writable: @wready<br>\n" if $debug;
  foreach $socket (@wready) {
    print $socket $msg;
  }

  my @rready = $s->can_read($timeout);
  print "DEBUG: readable: @rready<br>\n" if $debug;
  while(@rready = $s->can_read($timeout)) {
    foreach $socket (@rready) {
      $socket->sysread($buffer, 1024);
      $accumbuffer .= $buffer;
    }
  }

  $accumbuffer =~ s/\377\377\377\377print\n//g;
  my @response = split /^/m, $accumbuffer;
  shift @response;
  
  return \@response;
}

sub q3commandsingle {
  my $host = shift @_ || "";
  my $port = shift @_ || "";
  my $timeout = shift @_ || "";
  my $msg = shift @_ || "";
  my $buffer;
  my $code = "\377\377\377\377";
  $msg = $code.$msg;
 
  my $ok=1;
  my $sock=IO::Socket::INET->new(PeerAddr => $host,
                            PeerPort => $port,
                            Proto    => 'udp')
                or $ok=0;
  if ($ok)  {
  print $sock $msg;
  eval {
    local $SIG{ALRM} = sub { die "alarm\n" }; # NB: \n required
    alarm $timeout;
    $sock->sysread($buffer,1024);  
    alarm 0;
  };
  close($sock);
  my @response = split /^/m, $buffer;
  shift @response;
  
  return \@response;
  }
}
